/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib.daten;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeListDefinition;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DoubleAttributeType;
import de.bsvrz.dav.daf.main.config.IntegerAttributeType;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.dav.daf.main.config.ReferenceType;
import de.bsvrz.dav.daf.main.config.StringAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.TimeAttributeType;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Utility-Klasse zur Übersetzung von Datenverteiler-Datensätzen in
 * serialisierbare Parameterdatensätze und umgekehrt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public final class ParameterDatenFactory {

	/** kapselt Daten eines Parameters mit einem Änderungsstatus. */
	public static final class DataWithState {

		private final Data data;
		private final boolean changed;

		private DataWithState(final Data data, final boolean changed) {
			super();
			this.data = data;
			this.changed = changed;
		}

		/**
		 * liefert die Daten oder null.
		 *
		 * @return die Daten oder null
		 */
		public Data getData() {
			return data;
		}

		/**
		 * liefert den Änderungsstatus.
		 *
		 * @return den Status
		 */
		public boolean isChanged() {
			return changed;
		}
	}

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * füllt ein übergebenes Element eines Attributs aus einem
	 * Datenverteiler-Datensatz mit den Daten aus dem entsprechenden Attribut
	 * eines Parameter-Datensatzes.
	 *
	 * @param verbindung
	 *            die zu Grunde liegende Verbindung
	 * @param data
	 *            das Element des Datenverteiler-Datensatzes
	 * @param attribut
	 *            des Attribut des Parameter-Datensatzes
	 */
	private static boolean fuelleAttribut(final ClientDavInterface verbindung, final Data data,
			final Attribut<?> attribut) {

		boolean result = true;

		if (attribut instanceof Array) {
			final int size = ((Array) attribut).size();
			data.asArray().setLength(size);
			for (int idx = 0; idx < size; idx++) {
				final Attribut<?> neuesAttribut = ((Array) attribut).get(idx);
				result &= ParameterDatenFactory.fuelleAttribut(verbindung, data.asArray().getItem(idx), neuesAttribut);
			}

		} else if (attribut instanceof Attributliste) {
			for (final Attribut<?> att : ((Attributliste) attribut).get()) {
				result &= ParameterDatenFactory.fuelleAttribut(verbindung, data.getItem(att.getName()), att);
			}
		} else {
			final AttributeType type = data.getAttributeType();
			if (type instanceof StringAttributeType) {
				if (attribut instanceof Zeichenkette) {
					data.asTextValue().setText(((Zeichenkette) attribut).get());
				}
			} else if (type instanceof IntegerAttributeType) {
				if (attribut instanceof Ganzzahl) {
					data.asUnscaledValue().set(((Ganzzahl) attribut).get());
				}
			} else if (type instanceof DoubleAttributeType) {
				if (attribut instanceof Kommazahl) {
					data.asUnscaledValue().set(((Kommazahl) attribut).get());
				}
			} else if (type instanceof TimeAttributeType) {
				if (attribut instanceof Zeitstempel) {
					data.asTimeValue().setMillis(((Zeitstempel) attribut).get());
				}
			} else if (type instanceof ReferenceAttributeType) {
				if (attribut instanceof ObjektReferenz) {
					final Long objId = ((ObjektReferenz) attribut).get();

					if (objId == 0) {
						data.asReferenceValue().setSystemObject(null);
					} else {
						SystemObject sysObj = null;
						final String pid = ((ObjektReferenz) attribut).getPid();
						if (((ReferenceAttributeType) type).getReferenceType() == ReferenceType.ASSOCIATION) {
							if (pid == null) {
								ParameterDatenFactory.LOGGER
										.warning("Assoziative Referenz auf Object " + objId + " hat keine PID");
							} else {
								sysObj = verbindung.getDataModel().getObject(pid);
							}
						}

						if (sysObj == null) {
							sysObj = verbindung.getDataModel().getObject(objId);
						}

						if (sysObj == null) {
							ParameterDatenFactory.LOGGER
									.warning("Referenz auf Object " + objId + " oder PID: " + pid + " nicht gefunden");
						}

						data.asReferenceValue().setSystemObject(sysObj);
					}
				}
			}
		}

		return result;
	}

	/**
	 * wandelt einen Parameter-Datensatz in einem Datenverteiler-Datensatz um.
	 *
	 * @param verbindung
	 *            die zu Grunde liegende Verbindung
	 * @param atg
	 *            die Attributgruppe für die Zieldaten
	 * @param datensatz
	 *            der Parameter-Datensatz
	 * @return der Datenverteiler-Datensatz
	 */
	public static Data getData(final ClientDavInterface verbindung, final AttributeGroup atg,
			final Datensatz datensatz) {
		final DataWithState dataWithState = ParameterDatenFactory.getDataWithState(verbindung, atg, datensatz);
		if (dataWithState == null) {
			return null;
		}
		return dataWithState.getData();
	}

	/**
	 * wandelt einen Parameter-Datensatz in einem Datenverteiler-Datensatz um.
	 *
	 * @param verbindung
	 *            die zu Grunde liegende Verbindung
	 * @param atg
	 *            die Attributgruppe für die Zieldaten
	 * @param datensatz
	 *            der Parameter-Datensatz
	 * @return der Datenverteiler-Datensatz
	 */
	public static DataWithState getDataWithState(final ClientDavInterface verbindung, final AttributeGroup atg,
			final Datensatz datensatz) {

		AttributeGroup usedAtg = atg;

		if (null == datensatz || !datensatz.isValid()) {
			return null;
		}

		if (null == usedAtg) {
			usedAtg = verbindung.getDataModel().getAttributeGroup(datensatz.getName());
			if (null == usedAtg) {
				throw new IllegalArgumentException("Die im Datensatz definierte PID " + datensatz.getName()
						+ " beschreibt keine gültige Atributgruppe");
			}
		} else {
			if (!usedAtg.getPid().equals(datensatz.getName())) {
				throw new IllegalArgumentException("Ungültige Atributgruppe für Datensatz, Datensatz-ATG: "
						+ datensatz.getName() + ", Ziel: " + usedAtg.getPid());
			}
		}

		Data result = verbindung.createData(usedAtg);
		result.setToDefault();

		boolean changed = false;
		for (final Attribut<?> attribut : datensatz.get()) {
			final Data data = result.getItem(attribut.getName());
			changed = changed || !ParameterDatenFactory.fuelleAttribut(verbindung, data, attribut);
		}

		if (atg != null && "atg.parametrierung".equals(atg.getPid())) {
			if (!result.isDefined()) {
				ParameterDatenFactory.LOGGER.info("Bereinige leere Referenzen im Parameterdatensatz");
				result = ParameterDatenFactory.repairParameterDatensatzData(verbindung, atg, result);
				changed = true;
			}
		}

		return new DataWithState(result, changed);
	}

	private static Data repairParameterDatensatzData(final ClientDavInterface verbindung, final AttributeGroup atg,
			final Data result) {

		if (!"atg.parametrierung".equals(result.getName())) {
			return result;
		}

		final Data.Array parameterSaetze = result.getArray("ParameterSatz");
		final Data neueDaten = verbindung.createData(atg);

		neueDaten.getItem("Urlasser").getTextValue("Ursache")
				.setText("Nicht mehr auflösbare Referenzen aus dem Parameterdatensatz entfernt");

		final Data.Array neueParameterSaetze = neueDaten.getArray("ParameterSatz");
		neueParameterSaetze.setLength(ParameterDatenFactory.ermittleGueltigeParameterSaetze(parameterSaetze));

		int indexNeu = 0;
		for (int idx = 0; idx < parameterSaetze.getLength(); idx++) {
			if (ParameterDatenFactory.isParameterSatzReparierbar(parameterSaetze.getItem(idx))) {
				ParameterDatenFactory.repariereParameterDatenSatz(neueParameterSaetze.getItem(indexNeu),
						parameterSaetze.getItem(idx));
				indexNeu++;
			}
		}

		return neueDaten;
	}

	private static void repariereParameterDatenSatz(final Data neuDaten, final Data alteDaten) {

		final List<SystemObject> bereichsObjekte = new ArrayList<>();
		final de.bsvrz.dav.daf.main.Data.Array alteBereiche = alteDaten.getArray("Bereich");
		for (int idx = 0; idx < alteBereiche.getLength(); idx++) {
			final SystemObject systemObject = alteBereiche.getReferenceValue(idx).getSystemObject();
			if (systemObject != null) {
				bereichsObjekte.add(systemObject);
			}
		}

		final de.bsvrz.dav.daf.main.Data.Array neueBereiche = neuDaten.getArray("Bereich");
		neueBereiche.setLength(bereichsObjekte.size());
		int idx = 0;
		for (final SystemObject systemObject : bereichsObjekte) {
			neueBereiche.getReferenceValue(idx++).setSystemObject(systemObject);
		}

		final de.bsvrz.dav.daf.main.Data.Array alteDatenSpezifikationen = alteDaten.getArray("DatenSpezifikation");
		final de.bsvrz.dav.daf.main.Data.Array neueDatenSpezifikationen = neuDaten.getArray("DatenSpezifikation");
		neueDatenSpezifikationen
				.setLength(ParameterDatenFactory.ermittleGueltigeDatenSpezifikationen(alteDatenSpezifikationen));

		int indexNeu = 0;
		for (int dsIdx = 0; dsIdx < alteDatenSpezifikationen.getLength(); dsIdx++) {
			if (ParameterDatenFactory.isDatenSpezifikationReparierbar(alteDatenSpezifikationen.getItem(dsIdx))) {
				ParameterDatenFactory.repariereDatenSpezifikation(neueDatenSpezifikationen.getItem(indexNeu),
						alteDatenSpezifikationen.getItem(dsIdx));
				indexNeu++;
			}
		}

		neuDaten.getItem("Einstellungen").getUnscaledValue("Parametrieren")
				.set(alteDaten.getItem("Einstellungen").getUnscaledValue("Parametrieren").intValue());
	}

	private static void repariereDatenSpezifikation(final Data neuDaten, final Data alteDaten) {

		final List<SystemObject> objekte = new ArrayList<>();
		final de.bsvrz.dav.daf.main.Data.Array alteObjekte = alteDaten.getArray("Objekt");
		for (int idx = 0; idx < alteObjekte.getLength(); idx++) {
			final SystemObject systemObject = alteObjekte.getReferenceValue(idx).getSystemObject();
			if (systemObject != null) {
				objekte.add(systemObject);
			}
		}

		final de.bsvrz.dav.daf.main.Data.Array neueObjekte = neuDaten.getArray("Objekt");
		neueObjekte.setLength(objekte.size());
		int neuIdx = 0;
		for (final SystemObject systemObject : objekte) {
			neueObjekte.getReferenceValue(neuIdx++).setSystemObject(systemObject);
		}

		/* Attributgruppen. */

		final List<SystemObject> atgs = new ArrayList<>();
		final de.bsvrz.dav.daf.main.Data.Array alteAtgs = alteDaten.getArray("AttributGruppe");
		for (int idx = 0; idx < alteAtgs.getLength(); idx++) {
			final SystemObject systemObject = alteAtgs.getReferenceValue(idx).getSystemObject();
			atgs.add(systemObject);
		}

		final de.bsvrz.dav.daf.main.Data.Array neueAtgs = neuDaten.getArray("AttributGruppe");
		neueAtgs.setLength(atgs.size());
		neuIdx = 0;
		for (final SystemObject systemObject : atgs) {
			neueAtgs.getReferenceValue(neuIdx++).setSystemObject(systemObject);
		}

		neuDaten.getUnscaledValue("SimulationsVariante")
				.set(alteDaten.getUnscaledValue("SimulationsVariante").intValue());
	}

	private static int ermittleGueltigeParameterSaetze(final de.bsvrz.dav.daf.main.Data.Array parameterSaetze) {
		int count = 0;
		for (int idx = 0; idx < parameterSaetze.getLength(); idx++) {
			if (ParameterDatenFactory.isParameterSatzReparierbar(parameterSaetze.getItem(idx))) {
				count++;
			}
		}
		return count;
	}

	private static int ermittleGueltigeDatenSpezifikationen(final de.bsvrz.dav.daf.main.Data.Array spezifikationen) {
		int count = 0;
		for (int idx = 0; idx < spezifikationen.getLength(); idx++) {
			if (ParameterDatenFactory.isDatenSpezifikationReparierbar(spezifikationen.getItem(idx))) {
				count++;
			}
		}
		return count;
	}

	private static boolean isParameterSatzReparierbar(final Data item) {

		final de.bsvrz.dav.daf.main.Data.Array bereiche = item.getArray("Bereich");
		if (bereiche.getLength() > 0) {
			boolean reparierbar = false;
			for (int idx = 0; idx < bereiche.getLength(); idx++) {
				if (bereiche.getItem(idx).asReferenceValue().getSystemObject() != null) {
					reparierbar = true;
					break;
				}
			}

			if (!reparierbar) {
				return false;
			}
		}

		final de.bsvrz.dav.daf.main.Data.Array datenSpezifikationen = item.getArray("DatenSpezifikation");
		if (datenSpezifikationen.getLength() > 0) {
			boolean reparierbar = false;
			for (int idx = 0; idx < datenSpezifikationen.getLength(); idx++) {
				if (ParameterDatenFactory.isDatenSpezifikationReparierbar(datenSpezifikationen.getItem(idx))) {
					reparierbar = true;
					break;
				}
			}

			if (!reparierbar) {
				return false;
			}
		}

		return true;
	}

	private static boolean isDatenSpezifikationReparierbar(final Data item) {

		final de.bsvrz.dav.daf.main.Data.Array objekte = item.getArray("Objekt");
		if (objekte.getLength() > 0) {
			boolean reparierbar = false;
			for (int idx = 0; idx < objekte.getLength(); idx++) {
				if (objekte.getItem(idx).asReferenceValue().getSystemObject() != null) {
					reparierbar = true;
					break;
				}
			}

			if (!reparierbar) {
				return false;
			}
		}

		final de.bsvrz.dav.daf.main.Data.Array attributGruppen = item.getArray("AttributGruppe");
		if (attributGruppen.getLength() > 0) {
			boolean reparierbar = false;
			for (int idx = 0; idx < attributGruppen.getLength(); idx++) {
				if (attributGruppen.getItem(idx).asReferenceValue().getSystemObject() != null) {
					reparierbar = true;
					break;
				}
			}

			if (!reparierbar) {
				return false;
			}
		}

		return true;
	}

	/**
	 * erzeugt aus einem Datenverteiler-Datensatz einen serialisierbaren
	 * Parameter-Datensatz.
	 *
	 * @param verbindung
	 *            die zu Grunde liegende Datenverteilerverbindung
	 * @param data
	 *            die Daten, die umgewandelt werden sollen
	 * @return die umgewandelten Daten
	 */
	public static Attribut<? extends Object> getSerializable(final ClientDavInterface verbindung, final Data data) {

		Attribut<?> result = null;

		if (null == data || null == data.getAttributeType()) {
			final DataModel model = verbindung.getDataModel();
			if (null == data) {
				result = new Datensatz("EMPTY");
			} else {
				result = new Datensatz(data.getName());
				final AttributeGroup atg = model.getAttributeGroup(data.getName());
				for (final Attribute att : atg.getAttributes()) {
					((Datensatz) result)
							.add(ParameterDatenFactory.getSerializable(verbindung, data.getItem(att.getName())));
				}
			}
		} else if (data.isArray()) {
			result = new Array(data.getName());
			final Data.Array array = data.asArray();
			final int length = array.getLength();
			for (int idx = 0; idx < length; idx++) {
				((Array) result).add(ParameterDatenFactory.getSerializable(verbindung, array.getItem(idx)));
			}
		} else if (data.isList()) {
			final AttributeListDefinition attDef = (AttributeListDefinition) data.getAttributeType();
			result = new Attributliste(data.getName());
			for (final Attribute att : attDef.getAttributes()) {
				((Attributliste) result)
						.add(ParameterDatenFactory.getSerializable(verbindung, data.getItem(att.getName())));
			}
		} else {
			final AttributeType type = data.getAttributeType();
			if (type instanceof StringAttributeType) {
				result = new Zeichenkette(data.getName());
				((Zeichenkette) result).set(data.asTextValue().getText());
			} else if (type instanceof IntegerAttributeType) {
				result = new Ganzzahl(data.getName());
				((Ganzzahl) result).set(data.asUnscaledValue().longValue());
			} else if (type instanceof DoubleAttributeType) {
				result = new Kommazahl(data.getName());
				((Kommazahl) result).set(data.asUnscaledValue().doubleValue());
			} else if (type instanceof TimeAttributeType) {
				result = new Zeitstempel(data.getName());
				((Zeitstempel) result).set(data.asTimeValue().getMillis());
			} else if (type instanceof ReferenceAttributeType) {
				result = new ObjektReferenz(data.getName());
				final SystemObject ref = data.asReferenceValue().getSystemObject();
				if (null == ref) {
					((ObjektReferenz) result).set(0L);
				} else {
					((ObjektReferenz) result).set(ref.getId());
					if (((ReferenceAttributeType) type).getReferenceType() == ReferenceType.ASSOCIATION) {
						((ObjektReferenz) result).setPid(ref.getPid());
					}
				}
			}
		}

		return result;
	}

	/**
	 * privater Konstruktor, verhindert das Anlegen von INstanzen dieser Klasse.
	 */
	private ParameterDatenFactory() {
		super();
	}
}
